/* This file is Copyright 2000-2009 Meyer Sound Laboratories Inc.  See the included LICENSE.txt file for details. */  

#include <netdb.h>
#include <unistd.h>
#include <sys/socket.h>
#include <time.h>

#include "dataio/ChildProcessDataIO.h"
#include "iogateway/PlainTextMessageIOGateway.h"
#include "reflector/StorageReflectConstants.h"
#include "util/NetworkUtilityFunctions.h"
#include "system/SetupSystem.h"

using namespace muscle;

#define TEST(x) if ((x) != B_NO_ERROR) printf("Test failed, line %i\n",__LINE__)

static void PrintUsageAndExit()
{
   LogTime(MUSCLE_LOG_INFO, "Usage:  ./testchildprocess <count> <cmd> [args]\n");
   exit(10);
}

// This program is equivalent to the portableplaintext client, except
// that we communicate with a child process instead of a socket.
int main(int argc, char ** argv)
{
   CompleteSetupSystem css;

#ifdef SELECT_ON_FILE_DESCRIPTORS_NOT_AVAILABLE
   printf("Warning:  This program doesn't run very well on this OS, because the OS can't select() on stdin.  You'll need to press return a lot.\n");
#endif

   if (argc < 3) PrintUsageAndExit();

   uint32 numProcesses = atol(argv[1]);
   if (numProcesses == 0) PrintUsageAndExit();

   const char * cmd = argv[2];

   Queue<DataIORef> refs;
   for (uint32 i=0; i<numProcesses; i++)
   {
      ChildProcessDataIO * dio = new ChildProcessDataIO(false);
      refs.AddTail(DataIORef(dio));
      printf("About To Launch child process #"UINT32_FORMAT_SPEC":  [%s]\n", i+1, cmd); fflush(stdout);
      ConstSocketRef s = (dio->LaunchChildProcess(argc-2, ((const char **) argv)+2, true) == B_NO_ERROR) ? dio->GetSelectSocket() : ConstSocketRef();
      printf("Finished Launching child process #"UINT32_FORMAT_SPEC":  [%s]\n", i+1, cmd); fflush(stdout);
      if (s() == NULL)
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "Error launching child process #"UINT32_FORMAT_SPEC" [%s]!\n", i+1, cmd);
         return 10;
      }
   }

   for (uint32 i=0; i<refs.GetNumItems(); i++)
   {
      printf("------------ CHILD PROCESS #"UINT32_FORMAT_SPEC" ------------------\n", i+1);
      PlainTextMessageIOGateway gw;
      gw.SetDataIO(refs[i]);
      char text[1000] = "";
      fd_set readSet, writeSet;
      ConstSocketRef s = refs[i]()->GetSelectSocket();
      while(s())
      {
         FD_ZERO(&readSet);
         FD_ZERO(&writeSet);

         int fd = s.GetFileDescriptor();
         int maxfd = fd;
         FD_SET(fd, &readSet);
         struct timeval * timeout = NULL;
         struct timeval poll = {0, 0};

         if (gw.HasBytesToOutput()) FD_SET(fd, &writeSet);

#ifdef SELECT_ON_FILE_DESCRIPTORS_NOT_AVAILABLE
         // BeOS can't select() on stdin (yet), so just make the user press enter or whatever
         printf("Enter something => "); 
         fflush(stdout);
         fgets(text, sizeof(text), stdin);
         char * ret = strchr(text, '\n'); if (ret) *ret = '\0';
         timeout = &poll;  // prohibit blocking in select()
#else
         if (STDIN_FILENO > maxfd) maxfd = STDIN_FILENO;
         FD_SET(STDIN_FILENO, &readSet);
#endif

         QueueGatewayMessageReceiver inQueue;
         while(s()) 
         {
            if (select(maxfd+1, &readSet, &writeSet, NULL, timeout) < 0) printf("portablereflectclient: select() failed!\n");

            timeout = &poll;  // only block the first time (Linux) or not at all (BeOS)

#ifndef SELECT_ON_FILE_DESCRIPTORS_NOT_AVAILABLE
            if (FD_ISSET(STDIN_FILENO, &readSet)) 
            {
               fgets(text, sizeof(text), stdin);
               char * ret = strchr(text, '\n'); if (ret) *ret = '\0';
            }
#endif

            if (text[0])
            {
               printf("Sending: [%s]\n",text);
               MessageRef msg = GetMessageFromPool();
               msg()->AddString(PR_NAME_TEXT_LINE, text);
               gw.AddOutgoingMessage(msg);
               text[0] = '\0';
            }
      
            bool reading = FD_ISSET(fd, &readSet);
            bool writing = FD_ISSET(fd, &writeSet);
            bool writeError = ((writing)&&(gw.DoOutput() < 0));
            bool readError  = ((reading)&&(gw.DoInput(inQueue) < 0));
            if ((readError)||(writeError))
            {
               printf("Connection closed, exiting.\n");
               s.Reset();
            }

            MessageRef incoming;
            while(inQueue.RemoveHead(incoming) == B_NO_ERROR)
            {
               printf("Heard message from server:-----------------------------------\n");
               const char * inStr;
               for (int i=0; (incoming()->FindString(PR_NAME_TEXT_LINE, i, &inStr) == B_NO_ERROR); i++) printf("Line %i: [%s]\n", i, inStr);
              
               printf("-------------------------------------------------------------\n");
            }

            if ((reading == false)&&(writing == false)) break;

            FD_ZERO(&readSet);
            FD_ZERO(&writeSet);
            FD_SET(fd, &readSet);
            if (gw.HasBytesToOutput()) FD_SET(fd, &writeSet);
         }
      }

      if (gw.HasBytesToOutput())
      {
         printf("Waiting for all pending messages to be sent...\n");
         while((gw.HasBytesToOutput())&&(gw.DoOutput() >= 0)) {printf ("."); fflush(stdout);}
      }
   }
   printf("\n\nBye!\n");
   return 0;
}
