/* This file is Copyright 2000-2009 Meyer Sound Laboratories Inc.  See the included LICENSE.txt file for details. */  

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "message/Message.h"
#include "system/SetupSystem.h"
#include "util/Hashtable.h"
#include "util/MiscUtilityFunctions.h"
#include "util/String.h"
#include "util/StringTokenizer.h"
#include "util/TimeUtilityFunctions.h"

using namespace muscle;

int _state = 0;

static void bomb(const char * fmt, ...);
void bomb(const char * fmt, ...)
{
   va_list va;
   va_start(va, fmt);
   vprintf(fmt, va);
   va_end(va);
   LogTime(MUSCLE_LOG_CRITICALERROR, "EXITING DUE TO ERROR (state = %i)!\n", _state);
   ExitWithoutCleanup(10);
}

int MyCompareFunc(const int & s1, const int & s2, void *);
int MyCompareFunc(const int & s1, const int & s2, void *)
{
   return s1-s2;
}

static int DoInteractiveTest();
static int DoInteractiveTest()
{
   Hashtable<int, String> table;

   // Prepopulate the table, for convenience
   for (int i=9; i>=0; i--)
   {
      char buf[32];
      sprintf(buf, "%i", i);
      table.Put(i, buf);
   }

   // Let's see if this reverses it
   {
      HashtableIterator<int, String> rev(table);
      int next;
      while(rev.GetNextKey(next) == B_NO_ERROR) table.MoveToFront(next);
   }

   // turn on auto-sorting
   table.SetKeyCompareFunction(MyCompareFunc);
   table.SetAutoSortMode(table.AUTOSORT_BY_KEY);

   while(true)
   {
      HashtableIterator<int, String> iter(table);

      {
         bool first = true;
         printf("\nCurrent contents: ");
         int nextKey;
         while(iter.GetNextKey(nextKey) == B_NO_ERROR)
         {
            String * nextValue = iter.GetNextValue();
            MASSERT((atoi(nextValue->Cstr()) == nextKey), "value/key mismatch A!\n");
            printf("%s%i", first?"":", ", nextKey);
            first = false;
         }
         printf(" (size="UINT32_FORMAT_SPEC")\nEnter command: ", table.GetNumItems()); fflush(stdout);
      }

      char buf[512];
      if (fgets(buf, sizeof(buf), stdin))
      {
         StringTokenizer tok(buf, " ");
         const char * arg0 = tok();
         const char * arg1 = tok(); 
         const char * arg2 = tok(); 

         char command = arg0 ? arg0[0] : '\0';

         // For extra fun, let's put a half-way-done iterator here to see what happens
         printf("Concurrent: ");
         HashtableIterator<int, String> iter(table);
         bool first = true;
         int nextKey = 0;
         if (table.HasItems())
         {
            for (int i=table.GetNumItems()/2; i>=0; i--) 
            {
               MASSERT(iter.GetNextKey(nextKey) == B_NO_ERROR, "Not enough keys in table!?!?\n");
               String * nextValue = iter.GetNextValue();
               MASSERT((atoi(nextValue->Cstr()) == nextKey), "value/key mismatch B!\n");

               printf("%s%i", first?"":", ", nextKey);
               first = false;
            }
         }

         switch(command)
         { 
            case 'F':  
               if ((arg1)&&(arg2))
               {
                  printf("%s(%s %i before %i)", first?"":", ", (table.MoveToBefore(atoi(arg1),atoi(arg2)) == B_NO_ERROR) ? "Befored" : "FailedToBefored", atoi(arg1), atoi(arg2));
                  first = false;
               }
               else printf("(No arg1 or arg2!)");
            break;
            
            case 'B':  
               if ((arg1)&&(arg2))
               {
                  printf("%s(%s %i behind %i)", first?"":", ", (table.MoveToBehind(atoi(arg1),atoi(arg2)) == B_NO_ERROR) ? "Behinded" : "FailedToBehind", atoi(arg1), atoi(arg2));
                  first = false;
               }
               else printf("(No arg1 or arg2!)");
            break;
            
            case 'f':  
               if (arg1)
               {
                  printf("%s(%s %i)", first?"":", ", (table.MoveToFront(atoi(arg1)) == B_NO_ERROR) ? "Fronted" : "FailedToFront", atoi(arg1));
                  first = false;
               }
               else printf("(No arg1!)");
            break;
            
            case 'b':  
               if (arg1)
               {
                  printf("%s(%s %i)", first?"":", ", (table.MoveToBack(atoi(arg1)) == B_NO_ERROR) ? "Backed" : "FailedToBack", atoi(arg1));
                  first = false;
               }
               else printf("(No arg1!)");
            break;
            
            case 'p':  
               if (arg1)
               {
                  printf("%s(%s %i)", first?"":", ", (table.Put(atoi(arg1), arg1) == B_NO_ERROR) ? "Put" : "FailedToPut", atoi(arg1));
                  first = false;
               }
               else printf("(No arg1!)");
            break;
            
            case 'r':  
               if (arg1)
               {
                  printf("%s(%s %i)", first?"":", ", (table.Remove(atoi(arg1)) == B_NO_ERROR) ? "Removed" : "FailedToRemove", atoi(arg1));
                  first = false;
               }
               else printf("(No arg1!)");
            break;

            case 'c':  
               printf("%s(Clearing table)", first?"":", ");
               table.Clear();
               first = false;
            break;

            case 's':
               printf("%s(Sorting table)", first?"":", ");
               table.Sort();
               first = false;
            break;

            case 'q':  
               printf("%sQuitting\n", first?"":", ");
               first = false;
               return 0;
            break;
         }

         while(iter.GetNextKey(nextKey) == B_NO_ERROR)
         {
            String * nextValue = iter.GetNextValue();
            MASSERT((atoi(nextValue->Cstr()) == nextKey), "value/key mismatch C!\n");

            printf("%s%i", first?"":", ", nextKey);
            first = false;
         }
         printf("\n");
      }
      else return 0;
   }
}

static void CheckTable(Hashtable<int, bool> & table, uint32 numItems, bool backwards)
{
   uint32 count = 0;
   int last = backwards ? RAND_MAX : 0;
   int key  = 0;
   HashtableIterator<int, bool> iter(table,backwards?HTIT_FLAG_BACKWARDS:0);
   while(iter.GetNextKey(key) == B_NO_ERROR) 
   {
      if (backwards ? (last < key) : (last > key)) printf("ERROR!  Sort out of order in %s traversal!!!!\n", backwards?"backwards":"forwards");
      key = last;
      count++;
   }
   if (count != numItems) printf("ERROR!  Count is different!  "UINT32_FORMAT_SPEC" vs "UINT32_FORMAT_SPEC"\n", count, numItems);
}

static void TestIteratorSanityOnRemoval(bool backwards, bool useShortForm)
{
   LogTime(MUSCLE_LOG_INFO, "Testing iterator sanity (direction=%s) (form=%s)\n", backwards?"backwards":"forwards", useShortForm?"short":"classic");

   static const uint32 COUNT = 100;
   for (uint32 i=0; i<COUNT; i++)
   {
      Hashtable<int32, int32> table;
      for (uint32 j=0; j<COUNT; j++) table.Put(j, j+COUNT);

      uint32 numPairsFound = 0;
      int32 prevKey = backwards ? (int32)COUNT : -1;
      LogTime(MUSCLE_LOG_DEBUG, " Beginning traversal...\n");
      if (useShortForm)
      {
         for(HashtableIterator<int32,int32> iter(table, backwards?HTIT_FLAG_BACKWARDS:0); iter.HasMoreKeys(); iter++)
         {
            int32 expectedKey = (backwards?(prevKey-1):(prevKey+1));
            int32 gotKey      = iter.GetKey();
            int32 gotValue    = iter.GetValue();

            LogTime(MUSCLE_LOG_TRACE, "  Iter returned "INT32_FORMAT_SPEC" -> "INT32_FORMAT_SPEC"\n", iter.GetKey(), iter.GetValue());
            if (gotKey != expectedKey)
            {
               LogTime(MUSCLE_LOG_CRITICALERROR, "Expected key="INT32_FORMAT_SPEC", got key="INT32_FORMAT_SPEC" (value="UINT32_FORMAT_SPEC")\n", expectedKey, gotKey, gotValue);
               ExitWithoutCleanup(10);
            }

            if ((gotKey%(i+1))==0) {LogTime(MUSCLE_LOG_TRACE, "    -> Deleting key="INT32_FORMAT_SPEC"\n", gotKey); table.Remove(gotKey);}

            numPairsFound++;
            prevKey = gotKey;
         }
      }
      else
      {
         HashtableIterator<int32,int32> iter(table, backwards?HTIT_FLAG_BACKWARDS:0);
         int32 gotKey, gotValue;
         while(iter.GetNextKeyAndValue(gotKey, gotValue) == B_NO_ERROR)
         {
            int32 expectedKey = (backwards?(prevKey-1):(prevKey+1));

            LogTime(MUSCLE_LOG_TRACE, "  Iter returned "INT32_FORMAT_SPEC" -> "INT32_FORMAT_SPEC"\n", gotKey, gotValue);
            if (gotKey != expectedKey)
            {
               LogTime(MUSCLE_LOG_CRITICALERROR, "Expected key="INT32_FORMAT_SPEC", got key="INT32_FORMAT_SPEC" (value="UINT32_FORMAT_SPEC")\n", expectedKey, gotKey, gotValue);
               ExitWithoutCleanup(10);
            }

            if ((gotKey%(i+1))==0) {LogTime(MUSCLE_LOG_TRACE, "  -> Deleting key="INT32_FORMAT_SPEC"\n", gotKey); table.Remove(gotKey);}

            numPairsFound++;
            prevKey = gotKey;
         }
      }
      if (numPairsFound != COUNT)
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "Expected to iterate across "UINT32_FORMAT_SPEC" pairs, only saw "UINT32_FORMAT_SPEC"!\n", COUNT, numPairsFound);
         ExitWithoutCleanup(10);
      }
   }
}

// This program exercises the Hashtable class.
int main(int argc, char ** argv)
{
   CompleteSetupSystem css;
   Message temp; if (ParseArgs(argc, argv, temp) == B_NO_ERROR) HandleStandardDaemonArgs(temp);

   if (temp.HasName("inter")) return DoInteractiveTest();

   // Test iterator behaviour when deleting keys
   TestIteratorSanityOnRemoval(false, false);
   TestIteratorSanityOnRemoval(false, true);
   TestIteratorSanityOnRemoval(true,  false);
   TestIteratorSanityOnRemoval(true,  true);

   {
      LogTime(MUSCLE_LOG_DEBUG, "Preparing large table for sort...\n");

      const uint32 numItems = 100000;
      Hashtable<int, bool> table;
      for (uint32 i=0; i<numItems; i++) table.Put((int)rand(), true);
      uint32 actualNumItems = table.GetNumItems();  // may be smaller than numItems, due to duplicate values!

      LogTime(MUSCLE_LOG_DEBUG, "Sorting...\n");
      uint64 start = GetRunTime64();
      table.SortByKey(MyCompareFunc);
      uint64 end = GetRunTime64();

      LogTime(MUSCLE_LOG_INFO, "Time to sort "UINT32_FORMAT_SPEC" items: " UINT64_FORMAT_SPEC "ms\n", numItems, (end-start)/1000);

      // Check the resulting sorted table for correctness in both directions
      CheckTable(table, actualNumItems, false);
      CheckTable(table, actualNumItems, true);
   }

   // Test the sort algorithm for efficiency and correctness
   {
      LogTime(MUSCLE_LOG_DEBUG, "Preparing large table for sort...\n");

      const uint32 numItems = 100000;
      Hashtable<int, bool> table;
      for (uint32 i=0; i<numItems; i++) table.Put((int)rand(), true);
      uint32 actualNumItems = table.GetNumItems();  // may be smaller than numItems, due to duplicate values!

      LogTime(MUSCLE_LOG_DEBUG, "Sorting...\n");
      uint64 start = GetRunTime64();
      table.SortByKey(MyCompareFunc);
      uint64 end = GetRunTime64();

      LogTime(MUSCLE_LOG_INFO, "Time to sort "UINT32_FORMAT_SPEC" items: " UINT64_FORMAT_SPEC "ms\n", numItems, (end-start)/1000);

      // Check the resulting sorted table for correctness in both directions
      CheckTable(table, actualNumItems, false);
      CheckTable(table, actualNumItems, true);
   }

   Hashtable<String, String> table;
   {
      table.Put("Hello", "World");
      table.Put("Peanut Butter", "Jelly");
      table.Put("Ham", "Eggs");
      table.Put("Pork", "Beans");
      table.Put("Slash", "Dot");
      table.Put("Data", "Mining");  
      table.Put("TestDouble", "Play");
      table.Put("Abbot", "Costello");
      table.Put("Laurel", "Hardy");
      table.Put("Thick", "Thin");
      table.Put("Butter", "Parkay");
      table.Put("Total", "Carnage");
      table.Put("Summer", "Time");
      table.Put("Terrible", "Twos");

      if (table.GetNumItems() != 14)
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "String table has %i entries in it, expected 14!\n", table.GetNumItems());
         ExitWithoutCleanup(10);
      }

      {
         LogTime(MUSCLE_LOG_INFO, "Test partial backwards iteration\n");
         HashtableIterator<String, String> iter(table, "Slash", HTIT_FLAG_BACKWARDS);
         String key, val;
         while(iter.GetNextKeyAndValue(key, val) == B_NO_ERROR) LogTime(MUSCLE_LOG_INFO,"[%s] <-> [%s]\n", key(), val());
      }

      String lookup;
      if (table.Get(String("Hello"), lookup) == B_NO_ERROR) LogTime(MUSCLE_LOG_DEBUG, "Hello -> %s\n", lookup());
                                                       else bomb("Lookup 1 failed.\n");
      if (table.Get(String("Peanut Butter"), lookup) == B_NO_ERROR) LogTime(MUSCLE_LOG_DEBUG, "Peanut Butter -> %s\n", lookup());
                                                               else bomb("Lookup 2 failed.\n");


      HashtableIterator<String, String> st(table);
      String nextKeyString;
      String nextValueString;

      LogTime(MUSCLE_LOG_INFO, "Testing normal traveral with GetNextKey() and Get()\n");
      while(st.GetNextKey(nextKeyString) == B_NO_ERROR)
      {
         status_t ret = table.Get(nextKeyString, nextValueString);
         LogTime(MUSCLE_LOG_WARNING, "t1 = %i %s: %s -> %s\n", st.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKeyString(), nextValueString());
         if (ret != B_NO_ERROR) bomb("Couldn't find string!\n");
      }

      LogTime(MUSCLE_LOG_INFO, "Testing normal traveral with GetNextKey() and GetNextValue()\n");
      st = table.GetIterator();
      while(st.GetNextKey(nextKeyString) == B_NO_ERROR)
      {
         status_t ret = st.GetNextValue(nextValueString);
         LogTime(MUSCLE_LOG_INFO, "t2 = %i %s: %s -> %s\n", st.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKeyString(), nextValueString());
         if (ret != B_NO_ERROR) bomb("Couldn't find string!\n");
      }

      LogTime(MUSCLE_LOG_INFO, "Testing delete-as-you-go traveral\n");
      st = table.GetIterator();
      while(st.GetNextKey(nextKeyString) == B_NO_ERROR)
      {
         status_t ret = st.GetNextValue(nextValueString);
         LogTime(MUSCLE_LOG_INFO, "t3 = %i %s: %s -> %s (tableSize="UINT32_FORMAT_SPEC")\n", st.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKeyString(), nextValueString(), table.GetNumItems());
         if (table.Remove(nextKeyString) != B_NO_ERROR) bomb("Couldn't remove string!\n");
#if 0
         {
            HashtableIterator<String,String> st2(table);
            while(st2.GetNextKey(nextKeyString) == B_NO_ERROR)
            {
               status_t ret = st2.GetNextValue(nextValueString);
               printf("  tx = %i %s: %s -> %s\n", st2.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKeyString(), nextValueString());
            }
         }
#endif
      }

      printf("Testing another normal traversal with GetNextKey() and Get(), should be empty\n");
      while(st.GetNextKey(nextKeyString) == B_NO_ERROR)
      {
         status_t ret = table.Get(nextKeyString, nextValueString);
         printf("t1 = %i %s: %s -> %s\n", st.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKeyString(), nextValueString());
      }

      Hashtable<uint32, const char *> sillyTable;
      sillyTable.Put(15, "Fifteen");
      sillyTable.Put(100, "One Hundred");
      sillyTable.Put(150, "One Hundred and Fifty");
      sillyTable.Put(200, "Two Hundred");
      sillyTable.Put((uint32)-1, "2^32 - 1!");
      if (sillyTable.ContainsKey((uint32)-1) == false) bomb("large value failed!");

      // Test the GetNextKeyAndValue() methods
      {
         {
            uint32 nextKey;
            const char * nextValue;
            printf("GNKAV 1:\n");
            HashtableIterator<uint32, const char *> iter(sillyTable);
            while(iter.GetNextKeyAndValue(nextKey, nextValue) == B_NO_ERROR) printf(UINT32_FORMAT_SPEC"->%s\n", nextKey, nextValue);
         }
         {
            uint32 nextKey;
            const char ** nextValue;
            printf("GNKAV 2:\n");
            HashtableIterator<uint32, const char *> iter(sillyTable);
            while(iter.GetNextKeyAndValue(nextKey, nextValue) == B_NO_ERROR) printf(UINT32_FORMAT_SPEC"->%s\n", nextKey, *nextValue);
         }
         {
            const uint32 * nextKey;
            const char * nextValue;
            printf("GNKAV 3:\n");
            HashtableIterator<uint32, const char *> iter(sillyTable);
            while(iter.GetNextKeyAndValue(nextKey, nextValue) == B_NO_ERROR) printf(UINT32_FORMAT_SPEC"->%s\n", *nextKey, nextValue);
         }
         {
            const uint32 * nextKey;
            const char ** nextValue;
            printf("GNKAV 4:\n");
            HashtableIterator<uint32, const char *> iter(sillyTable);
            while(iter.GetNextKeyAndValue(nextKey, nextValue) == B_NO_ERROR) printf(UINT32_FORMAT_SPEC"->%s\n", *nextKey, *nextValue);
         }
      }

      const char * temp = NULL;
      sillyTable.Get(100, temp);
      sillyTable.Get(101, temp); // will fail
      printf("100 -> %s\n",temp);

      printf("Entries in sillyTable:\n");
      HashtableIterator<uint32, const char *> it(sillyTable);

      uint32 nextKey;
      const char * nextValue = NULL;
      while(it.GetNextKey(nextKey) == B_NO_ERROR)
      {
         status_t ret = sillyTable.Get(nextKey, nextValue);
         printf("%i %s: "UINT32_FORMAT_SPEC" -> %s\n", it.HasMoreKeys(), (ret == B_NO_ERROR) ? "OK" : "ERROR", nextKey, nextValue);
      }
      printf("XXX %i\n", it.HasMoreKeys());

      printf("... and by value:\n");
      while(it.GetNextValue(nextValue) == B_NO_ERROR) printf("%i value: %s\n", it.HasMoreValues(), nextValue);
      printf("YYY %i\n", it.HasMoreValues());
   }
   table.Clear();   

   printf("Begin torture test!\n");
   _state = 4;
   bool fastClear = false;
   {
      Hashtable<String, uint32> t;
      for (uint32 numEntries=1; numEntries < 1000; numEntries++)
      {
         uint32 half = numEntries/2;
         bool ok = true;

         printf(UINT32_FORMAT_SPEC" ", numEntries); fflush(stdout);
         _state = 5;
         {
            for(uint32 i=0; i<numEntries; i++)
            {
               char temp[300];
               sprintf(temp, UINT32_FORMAT_SPEC, i);
               if (t.Put(temp, i) != B_NO_ERROR)
               {
                  printf("Whoops, (hopefully simulated) memory failure!  (Put("UINT32_FORMAT_SPEC"/"UINT32_FORMAT_SPEC") failed) ... recovering\n", i, numEntries);

                  ok = false;
                  numEntries--;  // let's do this one over
                  half = i;    // so the remove code won't freak out about not everything being there
                  break;
               }
            }
         }

         if (ok)
         {
            //printf("Checking that all entries are still there...\n");
            _state = 6;
            {
               if (t.GetNumItems() != numEntries) bomb("ERROR, WRONG SIZE %i vs %i!\n", t.GetNumItems(), numEntries);
               for (int32 i=((int32)numEntries)-1; i>=0; i--)
               {
                  char temp[300];
                  sprintf(temp, UINT32_FORMAT_SPEC, i);
                  uint32 tv = 0;
                  if (t.Get(temp, tv) != B_NO_ERROR) bomb("ERROR, MISSING KEY [%s]\n", temp);
                  if (tv != ((uint32)i)) bomb("ERROR, WRONG KEY %s != "UINT32_FORMAT_SPEC"!\n", temp, tv);
               }
            }
      
            //printf("Iterating through table...\n");
            _state = 7;
            {
               uint32 count = 0;
               HashtableIterator<String, uint32> iter(t);
               String nextKey;
               while(iter.GetNextKey(nextKey) == B_NO_ERROR)
               {
                  char buf[300];
                  sprintf(buf, UINT32_FORMAT_SPEC, count);
                  if (nextKey != buf) bomb("ERROR:  iteration was wrong, item "UINT32_FORMAT_SPEC" was [%s] not [%s]!\n", count, nextKey(), buf);

                  uint32 temp = 0;
                  (void) iter.GetNextValue(temp);
                  if (temp != count) bomb("ERROR:  iteration value was wrong, item "UINT32_FORMAT_SPEC" was "UINT32_FORMAT_SPEC" not "UINT32_FORMAT_SPEC"!i!\n", count, temp, count);

                  count++;
               }
            }

            //printf("Removing the second half of the entries...\n");
            _state = 8;
            {
               for (uint32 i=half; i<numEntries; i++)
               {
                  char temp[64];
                  sprintf(temp, UINT32_FORMAT_SPEC, i);
                  uint32 tv = 0;  // just to shut the compiler up
                  if (t.Remove(temp, tv) != B_NO_ERROR) bomb("ERROR, MISSING REMOVE KEY [%s] A\n", temp);
                  if (tv != i) bomb("ERROR, REMOVE WAS WRONG VALUE "UINT32_FORMAT_SPEC"\n", tv);
               }
            }

            //printf("Iterating over only first half...\n");
            _state = 9;
            {
               uint32 sum = 0; 
               for (uint32 i=0; i<half; i++) sum += i;

               uint32 count = 0, checkSum = 0;
               HashtableIterator<String, uint32> iter(t);
               String nextKey;
               while(iter.GetNextKey(nextKey) == B_NO_ERROR) 
               {
                  count++;
                  uint32 temp = 0;
                  (void) iter.GetNextValue(temp);
                  checkSum += temp;
               }
               if (count != half) bomb("ERROR: Count mismatch "UINT32_FORMAT_SPEC" vs "UINT32_FORMAT_SPEC"!\n", count, numEntries);
               if (checkSum != sum)     bomb("ERROR: Sum mismatch "UINT32_FORMAT_SPEC" vs "UINT32_FORMAT_SPEC"!\n", sum, checkSum);
            }
         }

         //printf("Clearing Table (%s)\n", fastClear ? "Quickly" : "Slowly"); 
         _state = 10;
         if (fastClear) t.Clear();
         else
         {
            for (uint32 i=0; i<half; i++)
            {
               char temp[300];
               sprintf(temp, UINT32_FORMAT_SPEC, i);
               uint32 tv = 0;  // just to shut the compiler up
               if (t.Remove(temp, tv) != B_NO_ERROR) 
               {
                  String blahKey;
                  uint32 blah;
                  HashtableIterator<String, uint32> p(t);
                  while(p.GetNextKey(blahKey) == B_NO_ERROR)
                  {
                     if (p.GetNextValue(blah) == B_NO_ERROR) printf("[%s]->"UINT32_FORMAT_SPEC, blahKey(), blah);
                                                        else printf("[%s]->(no value!)\n", blahKey());
                  }
                  printf("\n");
                  bomb("ERROR, MISSING REMOVE KEY [%s] ("UINT32_FORMAT_SPEC"/"UINT32_FORMAT_SPEC") B\n", temp, i, half);
               }
               if (tv != i) bomb("ERROR, REMOVE WAS WRONG VALUE "UINT32_FORMAT_SPEC"\n", tv);
            }
         }

         HashtableIterator<String, uint32> paranoia(t);
         if (paranoia.GetNextKey()) bomb("ERROR, ITERATOR CONTAINED ITEMS AFTER CLEAR!\n");

         if (t.HasItems()) bomb("ERROR, SIZE WAS NON-ZERO ("UINT32_FORMAT_SPEC") AFTER CLEAR!\n", t.GetNumItems());
         fastClear = !fastClear;
      }
      printf("Finished torture test successfully, exiting.\n");
   }
}
